@@ -65,6 +65,7 @@ extern int __REDIRECT_NTH (__ttyname_r_alias, (int __fd, char *__buf,
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
 #include <sys/inotify.h>
+#include <sys/mount.h>
 #include <sys/socket.h>
 #include <sys/vfs.h>
 #include <sys/xattr.h>
@@ -141,9 +142,10 @@ dev_of_fd(int fd)
 {
     struct stat st;
     int ret, orig_errno;
+    libc_func(fstat, int, int, struct stat*);
 
     orig_errno = errno;
-    ret = fstat(fd, &st);
+    ret = _fstat(fd, &st);
     errno = orig_errno;
     if (ret < 0)
 	return 0;
@@ -342,6 +344,88 @@ is_emulated_device(const char *path, const mode_t st_mode)
     return !S_ISDIR(st_mode);
 }
 
+/* dirfd helper (openat family): intercept opening /dev and /sys from the root dir */
+static const char *
+resolve_dirfd_path(int dirfd, const char *pathname)
+{
+    libc_func(readlink, ssize_t, const char*, char *, size_t);
+    const char *p = NULL;
+    int trapped = 0;
+
+    if ((strncmp(pathname, "sys", 3) == 0 || strncmp(pathname, "dev", 3) == 0) &&
+        (pathname[3] == '/' || pathname[3] == '\0')) {
+        static char buf[PATH_MAX], link[PATH_MAX];
+        snprintf(buf, sizeof(buf), "/proc/self/fd/%d", dirfd);
+        if (_readlink(buf, link, sizeof(link)) == 1 && link[0] == '/') {
+            trapped = 1;
+            strncpy(link + 1, pathname, sizeof(link) - 2);
+            link[sizeof(link) - 1] = 0;
+            p = trap_path(link);
+        }
+    }
+
+    if (!trapped)
+        p = trap_path(pathname);
+
+    return p;
+}
+
+static bool is_dir_or_contained(const char *path, const char *dir, const char *subdir)
+{
+    if (!path || !dir)
+	return false;
+
+    const ssize_t subdir_len = strlen(subdir);
+    const size_t dir_len = strlen(dir);
+
+    return (dir_len + subdir_len <= strlen(path) &&
+	    strncmp(path, dir, dir_len) == 0 &&
+	    strncmp(path + dir_len, subdir, subdir_len) == 0 &&
+	    (path[dir_len + subdir_len] == '\0' || path[dir_len + subdir_len] == '/'));
+}
+
+static bool get_fd_path(int fd, char *path_out, size_t path_size)
+{
+    static char fdpath[PATH_MAX];
+    libc_func(readlink, ssize_t, const char*, char *, size_t);
+
+    snprintf(fdpath, sizeof fdpath, "/proc/self/fd/%i", fd);
+    int orig_errno = errno;
+    ssize_t linklen = _readlink(fdpath, path_out, path_size);
+    errno = orig_errno;
+    if (linklen < 0 || linklen >= (ssize_t)path_size)
+        return false;
+    path_out[linklen] = '\0';
+    return true;
+}
+
+static bool is_fd_in_mock(int fd, const char *subdir)
+{
+    static char linkpath[PATH_MAX];
+
+    if (!get_fd_path(fd, linkpath, sizeof linkpath)) {
+	perror("umockdev: failed to map fd to a path");
+	return false;
+    }
+
+    return is_dir_or_contained(linkpath, getenv("UMOCKDEV_DIR"), subdir);
+}
+
+/* Helper to adjust stat for known emulated devices */
+static inline void
+adjust_emulated_device_mode_rdev(const char *dev_path, mode_t *st_mode, dev_t *st_rdev)
+{
+    *st_mode &= ~S_IFREG;
+    if (*st_mode & S_ISVTX) {
+        *st_mode = S_IFBLK | (*st_mode & ~S_IFMT);
+        DBG(DBG_PATH, "  %s is an emulated block device\n", dev_path);
+    } else {
+        *st_mode = S_IFCHR | (*st_mode & ~S_IFMT);
+        DBG(DBG_PATH, "  %s is an emulated char device\n", dev_path);
+    }
+    *st_rdev = get_rdev(dev_path + 5);
+}
+
 /********************************
  *
  * fd -> pointer map
@@ -506,6 +590,40 @@ struct ioctl_fd_info {
     pthread_mutex_t sock_lock;
 };
 
+/* Helper to adjust fstat results for emulated devices */
+static void
+fstat_adjust_emulated_device(int fd, mode_t *st_mode, dev_t *st_rdev)
+{
+    /* Check if this fd is for an emulated device using the ioctl tracking system */
+    struct ioctl_fd_info *fdinfo;
+    if (fd_map_get(&ioctl_wrapped_fds, fd, (const void **)&fdinfo) && fdinfo->is_emulated) {
+        DBG(DBG_PATH, "fstat(%i): adjusting emulated device %s\n", fd, fdinfo->dev_path);
+        adjust_emulated_device_mode_rdev(fdinfo->dev_path, st_mode, st_rdev);
+        return;
+    }
+
+        /* Check if this untracked fd points to an emulated device in the mock testbed */
+    if (is_fd_in_mock(fd, "/dev")) {
+        static char linkpath[PATH_MAX];
+
+        if (get_fd_path(fd, linkpath, sizeof linkpath)) {
+            /* Extract the /dev/... part from the mock path */
+            const char *umockdev_dir = getenv("UMOCKDEV_DIR");
+            if (umockdev_dir) {
+                size_t prefix_len = strlen(umockdev_dir);
+                if (strncmp(linkpath, umockdev_dir, prefix_len) == 0 &&
+                    strncmp(linkpath + prefix_len, "/dev/", 5) == 0) {
+                    const char *dev_path = linkpath + prefix_len;
+                    if (is_emulated_device(linkpath, *st_mode)) {
+                        DBG(DBG_PATH, "fstat(%i): adjusting untracked emulated device %s\n", fd, dev_path);
+                        adjust_emulated_device_mode_rdev(dev_path, st_mode, st_rdev);
+                    }
+                }
+            }
+        }
+    }
+}
+
 static void
 ioctl_emulate_open(int fd, const char *dev_path, bool is_emulated)
 {
@@ -1230,18 +1348,9 @@ rettype name(const char *path, arg2t arg2, arg3t arg3, arg4t arg4) \
 }
 
 #define STAT_ADJUST_MODE \
-    if (ret == 0 && p != path && strncmp(path, "/dev/", 5) == 0			\
-	&& is_emulated_device(p, st->st_mode)) {				\
-	st->st_mode &= ~S_IFREG;						\
-	if (st->st_mode & S_ISVTX) {						\
-            st->st_mode = S_IFBLK | (st->st_mode & ~S_IFMT);			\
-	    DBG(DBG_PATH, "  %s is an emulated block device\n", path);		\
-	} else {								\
-            st->st_mode = S_IFCHR | (st->st_mode & ~S_IFMT);			\
-	    DBG(DBG_PATH, "  %s is an emulated char device\n", path);		\
-	}									\
-	st->st_rdev = get_rdev(path + 5);					\
-    }										\
+    if (ret == 0 && p != path && strncmp(path, "/dev/", 5) == 0 \
+	    && is_emulated_device(p, st->st_mode)) \
+        adjust_emulated_device_mode_rdev(path, &st->st_mode, &st->st_rdev);
 
 /* wrapper template for stat family; note that we abuse the sticky bit in
  * the emulated /dev to indicate a block device (the sticky bit has no
@@ -1277,7 +1386,7 @@ int prefix ## fstatat ## suffix (int dirfd, const char *path, struct stat ## suf
     libc_func(prefix ## fstatat ## suffix, int, int, const char*, struct stat ## suffix *, int); \
     int ret;									\
     TRAP_PATH_LOCK;								\
-    p = trap_path(path);							\
+    p = resolve_dirfd_path(dirfd, path);					\
     if (p == NULL) {								\
 	TRAP_PATH_UNLOCK;							\
 	return -1;								\
@@ -1322,7 +1431,7 @@ int prefix ## fxstatat ## suffix (int ver, int dirfd, const char *path, struct s
     libc_func(prefix ## fxstatat ## suffix, int, int, int, const char*, struct stat ## suffix *, int); \
     int ret;									\
     TRAP_PATH_LOCK;								\
-    p = trap_path(path);							\
+    p = resolve_dirfd_path(dirfd, path);					\
     if (p == NULL) {								\
 	TRAP_PATH_UNLOCK;							\
 	return -1;								\
@@ -1334,7 +1443,18 @@ int prefix ## fxstatat ## suffix (int ver, int dirfd, const char *path, struct s
     return ret;									\
 }
 
-
+/* wrapper template for fstat family */
+#define WRAP_FSTAT(prefix, suffix) \
+extern int prefix ## fstat ## suffix (int fd, struct stat ## suffix *st); \
+int prefix ## fstat ## suffix (int fd, struct stat ## suffix *st) \
+{ \
+    libc_func(prefix ## fstat ## suffix, int, int, struct stat ## suffix *); \
+    int ret = _ ## prefix ## fstat ## suffix(fd, st); \
+    if (ret == 0) { \
+        fstat_adjust_emulated_device(fd, &st->st_mode, &st->st_rdev); \
+    } \
+    return ret; \
+}
 /* wrapper template for open family */
 #define WRAP_OPEN(prefix, suffix) \
 int prefix ## open ## suffix (const char *path, int flags, ...)	    \
@@ -1422,11 +1542,13 @@ WRAP_2ARGS(int, -1, access, int);
 WRAP_STAT(,);
 WRAP_STAT(l,);
 WRAP_FSTATAT(,);
+WRAP_FSTAT(,);
 
 #ifdef __GLIBC__
 WRAP_STAT(,64);
 WRAP_STAT(l,64);
 WRAP_FSTATAT(,64);
+WRAP_FSTAT(,64);
 WRAP_FOPEN(,64);
 #if defined(__USE_FILE_OFFSET64) && defined(__USE_TIME64_REDIRECTS)
 #define stat64_time64 stat64
@@ -1469,60 +1591,26 @@ int statx(int dirfd, const char *pathname, int flags, unsigned mask, struct stat
 
     if (r == 0 && p != pathname && strncmp(pathname, "/dev/", 5) == 0
             && is_emulated_device(p, stx->stx_mode)) {
-        if (stx->stx_mode & S_ISVTX) {
-            stx->stx_mode = S_IFBLK | (stx->stx_mode & ~S_IFMT);
-            DBG(DBG_PATH, "  %s is an emulated block device (statx)\n", pathname);
-        } else {
-            stx->stx_mode = S_IFCHR | (stx->stx_mode & ~S_IFMT);
-            DBG(DBG_PATH, "  %s is an emulated char device (statx)\n", pathname);
-        }
-	unsigned maj, min;
-	if (get_rdev_maj_min(pathname + 5, &maj, &min)) {
-	    stx->stx_rdev_major = maj;
-	    stx->stx_rdev_minor = min;
-	} else {
-	    stx->stx_rdev_major = stx->stx_rdev_minor = 0;
-	}
+        /* Adjust mode using the common helper - need to handle type conversion for statx */
+        mode_t adjusted_mode = stx->stx_mode;
+        dev_t dummy_rdev = 0;
+        adjust_emulated_device_mode_rdev(pathname, &adjusted_mode, &dummy_rdev);
+        stx->stx_mode = adjusted_mode;
 
+        /* Handle statx-specific rdev fields */
+        unsigned maj, min;
+        if (get_rdev_maj_min(pathname + 5, &maj, &min)) {
+            stx->stx_rdev_major = maj;
+            stx->stx_rdev_minor = min;
+        } else {
+            stx->stx_rdev_major = stx->stx_rdev_minor = 0;
+        }
     }
     return r;
 }
 
 #endif /* __GLIBC__ */
 
-static bool is_dir_or_contained(const char *path, const char *dir, const char *subdir)
-{
-    if (!path || !dir)
-	return false;
-
-    const ssize_t subdir_len = strlen(subdir);
-    const size_t dir_len = strlen(dir);
-
-    return (dir_len + subdir_len <= strlen(path) &&
-	    strncmp(path, dir, dir_len) == 0 &&
-	    strncmp(path + dir_len, subdir, subdir_len) == 0 &&
-	    (path[dir_len + subdir_len] == '\0' || path[dir_len + subdir_len] == '/'));
-}
-
-static bool is_fd_in_mock(int fd, const char *subdir)
-{
-    static char fdpath[PATH_MAX];
-    static char linkpath[PATH_MAX];
-    libc_func(readlink, ssize_t, const char*, char *, size_t);
-
-    snprintf(fdpath, sizeof fdpath, "/proc/self/fd/%i", fd);
-    int orig_errno = errno;
-    ssize_t linklen = _readlink(fdpath, linkpath, sizeof linkpath);
-    errno = orig_errno;
-    if (linklen < 0 || linklen >= sizeof linkpath) {
-	perror("umockdev: failed to map fd to a path");
-	return false;
-    }
-    linkpath[linklen] = '\0';
-
-    return is_dir_or_contained(linkpath, getenv("UMOCKDEV_DIR"), subdir);
-}
-
 #define WRAP_FSTATFS(suffix) \
 int fstatfs ## suffix(int fd, struct statfs ## suffix *buf)	\
 { \
@@ -1570,29 +1658,15 @@ int __open64_2(const char *path, int flags);
 WRAP_OPEN(,);
 WRAP_OPEN2(__,_2);
 
-/* wrapper template for openat family; intercept opening /sys from the root dir */
+/* wrapper template for openat family */
 #define WRAP_OPENAT(prefix, suffix) \
 int prefix ## openat ## suffix (int dirfd, const char *pathname, int flags, ...)		\
 { \
     const char *p = NULL;									\
     libc_func(prefix ## openat ## suffix, int, int, const char *, int, ...);			\
-    libc_func(readlink, ssize_t, const char*, char *, size_t);					\
-    int trapped = 0, ret;									\
+    int ret;											\
     TRAP_PATH_LOCK;										\
-  \
-    if (strncmp(pathname, "sys", 3) == 0 && (pathname[3] == '/' || pathname[3] == '\0')) {	\
-        static char buf[PATH_MAX],link[PATH_MAX];						\
-        snprintf(buf, sizeof(buf), "/proc/self/fd/%d", dirfd);					\
-        if (_readlink(buf, link, sizeof(link)) == 1 && link[0] == '/') {			\
-            trapped = 1;									\
-            strncpy(link + 1, pathname, sizeof(link) - 2);					\
-            buf[sizeof(link) - 1] = 0;								\
-            p = trap_path(link);								\
-        } \
-    } \
-  \
-    if (!trapped)										\
-        p = trap_path(pathname);								\
+    p = resolve_dirfd_path(dirfd, pathname);							\
     if (p == NULL) { TRAP_PATH_UNLOCK; return -1; }						\
     DBG(DBG_PATH, "testbed wrapped " #prefix "openat" #suffix "(%s) -> %s\n", pathname, p);	\
     if (flags & (O_CREAT | O_TMPFILE)) {							\
@@ -1616,6 +1690,27 @@ WRAP_OPEN2(__,64_2);
 WRAP_OPENAT(, 64);
 #endif
 
+#if defined(__GLIBC__) && defined(HAVE_OPEN_TREE)
+int
+open_tree(int dfd, const char *filename, unsigned int flags)
+{
+    const char *p;
+    libc_func(open_tree, int, int, const char *, unsigned int);
+    int ret;
+
+    TRAP_PATH_LOCK;
+    p = resolve_dirfd_path(dfd, filename);
+    if (p == NULL) {
+        TRAP_PATH_UNLOCK;
+        return -1;
+    }
+    DBG(DBG_PATH, "testbed wrapped open_tree(%i, %s) -> %s\n", dfd, filename, p);
+    ret = _open_tree(dfd, p, flags);
+    TRAP_PATH_UNLOCK;
+    return ret;
+}
+#endif
+
 int
 inotify_add_watch(int fd, const char *path, uint32_t mask)
 {
